import java.io.*;

/**
 * Bitmap-Grafik (Microsoft-Windows)
 * mit 24 Bit per Pixel.
 * 
 * @author Heiner Kcker
 */
public class Bitmap24
{
    public static final int WHITE = 16777215; // entspricht 2^24 - 1

    public static final int BLACK = 0;

    private int[][] iClArr;

    /**
     * Konstruktor mit Grenangaben
     */
    public Bitmap24(int iPaRows, int iPaCols)
    {
        iClArr = new int[iPaRows][iPaCols];

        // Array mit Nullen setzen ( mal testen, ob berhaupt ntig )
        setAllPix(255);
    } // end Konstruktor

    /**
     * Konstruktor mit Grenangaben
     */
    public Bitmap24(int iPaRows, int iPaCols, final int pColor)
    {
        iClArr = new int[iPaRows][iPaCols];

        // Array mit Nullen setzen ( mal testen, ob berhaupt ntig )
        setAllPix(pColor);
    } // end Konstruktor

    /**
     * Setzen aller Pixel auf die festgelegte Farbe.
     *
     */
    public void setAllPix(final int pColor)
    {
        for (int iRow = 0; iRow < iClArr.length; iRow++)
        {
            for (int jCol = 0; jCol < iClArr[0].length; jCol++)
            {
                iClArr[iRow][jCol] = pColor;
            }
        }
    }

    /**
     * Setzen eines Pixels mit einer bestimmten Farbe
     */
    public void setpix(int iPaRow, int iPaCol, int iPaColor)
    {
        if (iPaRow >= 0 && iPaRow < iClArr.length)
        {
            if (iPaCol >= 0 && iPaCol < iClArr[iPaRow].length)
            {
                iClArr[iPaRow][iPaCol] = iPaColor;
            }
        }
    } // end method setpix

    /**
     * Linie zeichnen
     */
    public void drawLine(
        int iPaRow1,
        int iPaCol1,
        int iPaRow2,
        int iPaCol2,
        int iPaColor)
    {
        //hier noch Prfung Parameter
        if (iPaRow1 == iPaRow2)
        {
            // horizontale Linie
            //System.out.println("horizontale Linie");
            if (iPaCol1 > iPaCol2)
                // Punkte mssen getauscht werden, sonst klappt es nicht
            {
                int iTmp;
                iTmp = iPaRow1;
                iPaRow1 = iPaRow2;
                iPaRow2 = iTmp;
                iTmp = iPaCol1;
                iPaCol1 = iPaCol2;
                iPaCol2 = iTmp;
            }

            for (int i = iPaCol1; i <= iPaCol2; i++)
            {
                //System.out.println("Zeile " + iPaRow1 + " Spalte " + i);
                iClArr[iPaRow1][i] = iPaColor; // schneller als ber setpix
            }
        }
        else if (iPaCol1 == iPaCol2)
        {
            // vertikale Linie
            //System.out.println("vertikale Linie");
            if (iPaRow1 > iPaRow2)
                // Punkte mssen getauscht werden, sonst klappt es nicht
            {
                int iTmp;
                iTmp = iPaRow1;
                iPaRow1 = iPaRow2;
                iPaRow2 = iTmp;
                iTmp = iPaCol1;
                iPaCol1 = iPaCol2;
                iPaCol2 = iTmp;
            }

            for (int i = iPaRow1; i <= iPaRow2; i++)
            {
                iClArr[i][iPaCol1] = iPaColor; // schneller als ber setpix
            }
        }
        else if ((iPaRow2 - iPaRow1) < (iPaCol2 - iPaCol1))
        {
            //System.out.println("flache Linie");
            // flache Linie
            if (iPaCol1 > iPaCol2)
                // Punkte mssen getauscht werden, sonst klappt es nicht
            {
                int iTmp;
                iTmp = iPaRow1;
                iPaRow1 = iPaRow2;
                iPaRow2 = iTmp;
                iTmp = iPaCol1;
                iPaCol1 = iPaCol2;
                iPaCol2 = iTmp;
            }

            float iStep =
                ((float) iPaRow2 - (float) iPaRow1)
                    / ((float) iPaCol2 - (float) iPaCol1);
            // hier noch optmieren iStep gleich im if zuweisen
            float iRow = iPaRow1;
            for (int i = iPaCol1; i <= iPaCol2; i++)
            {
                iClArr[Math.round(iRow)][i] = iPaColor;
                // schneller als ber setpix
                iRow += iStep;
            }
        }
        else
        {
            // steile Linie
            //System.out.println("steile Linie");

            if (iPaRow1 > iPaRow2)
                // Punkte mssen getauscht werden, sonst klappt es nicht
            {
                int iTmp;
                iTmp = iPaRow1;
                iPaRow1 = iPaRow2;
                iPaRow2 = iTmp;
                iTmp = iPaCol1;
                iPaCol1 = iPaCol2;
                iPaCol2 = iTmp;
            }

            float iStep =
                (((float) iPaCol2) - ((float) iPaCol1))
                    / (((float) iPaRow2) - ((float) iPaRow1));
            float iCol = iPaCol1;
            for (int i = iPaRow1; i <= iPaRow2; i++)
            {
                //iClArr[ i ][ iCol ] = iPaColor ; // schneller als ber setpix
                setpix(i, Math.round(iCol), iPaColor);
                //System.out.println( "Zeile " + i + " Spalte " + Math.round(iCol));

                iCol += iStep;
            }
        }
    } // end method drawLine

    /**
     * Kreis Zeichnen
     */
    public void drawCircle(int iPaRow, int iPaCol, int iPaRadius, int iPaColor)
    {
        //float f1 ;
        float f2;
        float fTmp;

        fTmp = (float) iPaRadius * (float) iPaRadius; // Quadrat

        for (int i = 0; i <= iPaRadius; i++)
        {
            f2 = (float) Math.sqrt(fTmp - ((float) i * (float) i));
            setpix(iPaRow + Math.round(f2), iPaCol + i, iPaColor);
            setpix(iPaRow + Math.round(f2), iPaCol - i, iPaColor);
            setpix(iPaRow - Math.round(f2), iPaCol + i, iPaColor);
            setpix(iPaRow - Math.round(f2), iPaCol - i, iPaColor);

            f2 = (float) Math.sqrt(fTmp - ((float) i * (float) i));
            setpix(iPaRow + i, iPaCol + Math.round(f2), iPaColor);
            setpix(iPaRow + i, iPaCol - Math.round(f2), iPaColor);
            setpix(iPaRow - i, iPaCol + Math.round(f2), iPaColor);
            setpix(iPaRow - i, iPaCol - Math.round(f2), iPaColor);
        }
    } // end method drawCircle

    /**
     * Ausgabe des Pixel-Arrays in einen OutputStream 
     * in 24 Bit per Pixel.
     * RLE-Komprimierung wird in Bitmap-Fateien bei 
     * 24 Bit per Pixel nicht untersttzt.
     */
    public void writeBmpToStream(final OutputStream osPaStream) throws IOException
    {
        // File ID "BM" Dateikennung
        osPaStream.write('B');
        osPaStream.write('M');

        final int iBitmapWidth = (iClArr.length * (iClArr[0].length + 2) * 3);

        // Dateilnge
        BitmapUtil.writeInt32Intel(osPaStream, 54 + iBitmapWidth);

        // vier reservierte Bytes
        osPaStream.write((byte) 0);
        osPaStream.write((byte) 0);
        osPaStream.write((byte) 0);
        osPaStream.write((byte) 0);

        // Startadresse Bitmap 54 entspricht 36H ( keine Palette )
        BitmapUtil.writeInt32Intel(osPaStream, 54);
        //writeInt32Intel( osPaStream , 118 ) ; // Startadresse Bitmap

        // Groesse des Kopfes in Bytes, Position 0EH, Wert ist immer 40
        BitmapUtil.writeInt32Intel(osPaStream, 40);

        // Breite der Bitmap in Bildschirmpunkten
        BitmapUtil.writeInt32Intel(osPaStream, iClArr[0].length);

        // Hhe der Bitmap in Bildschirmpunkten
        BitmapUtil.writeInt32Intel(osPaStream, iClArr.length);

        // Anzahl der Farbebenen, muss 1 sein
        osPaStream.write((byte) 1);
        osPaStream.write((byte) 0);

        // Bits pro Pixel
        osPaStream.write((byte) 24);
        osPaStream.write((byte) 0);

        // Kennzeichnung fuer die Komprimierung ( jetzt noch null / keine Komprimierung )
        osPaStream.write((byte) 0);
        osPaStream.write((byte) 0);
        osPaStream.write((byte) 0);
        osPaStream.write((byte) 0);

        // Groesse des Bitmaps in Bytes
        BitmapUtil.writeInt32Intel(osPaStream, iBitmapWidth);

        // horizontale Aufloesung
        BitmapUtil.writeInt32Intel(osPaStream, 3780);

        // vertikale Aufloesung
        BitmapUtil.writeInt32Intel(osPaStream, 3780);

        // Anzahl der Farben aus der Farbtabelle
        BitmapUtil.writeInt32Intel(osPaStream, 0);

        // Anzahl wichtiger Farben  aus der Farbtabelle
        BitmapUtil.writeInt32Intel(osPaStream, 0);

        /* Palette entfaellt
              // Palette fuer 16 Farben
              // 00 schwarz
              osPaStream.write( (byte) 0 ) ; // Intensitaet blau
              osPaStream.write( (byte) 0 ) ; // Intensitaet gruen
              osPaStream.write( (byte) 0 ) ; // Intensitaet rot
              osPaStream.write( (byte) 0 ) ; // reserviert
        
              // 01 blau
              osPaStream.write( (byte) 126 ) ; // Intensitaet blau
              osPaStream.write( (byte) 0 ) ; // Intensitaet gruen
              osPaStream.write( (byte) 0 ) ; // Intensitaet rot
              osPaStream.write( (byte) 0 ) ; // reserviert
        
              // 02 gruen
              osPaStream.write( (byte) 0 ) ; // Intensitaet blau
              osPaStream.write( (byte) 126 ) ; // Intensitaet gruen
              osPaStream.write( (byte) 0 ) ; // Intensitaet rot
              osPaStream.write( (byte) 0 ) ; // reserviert
        
              // 03 rot
              osPaStream.write( (byte) 0 ) ; // Intensitaet blau
              osPaStream.write( (byte) 0 ) ; // Intensitaet gruen
              osPaStream.write( (byte) 126 ) ; // Intensitaet rot
              osPaStream.write( (byte) 0 ) ; // reserviert
        
              // 04 schwarz
              osPaStream.write( (byte) 0 ) ; // Intensitaet blau
              osPaStream.write( (byte) 0 ) ; // Intensitaet gruen
              osPaStream.write( (byte) 0 ) ; // Intensitaet rot
              osPaStream.write( (byte) 0 ) ; // reserviert
        
              // 05 schwarz
              osPaStream.write( (byte) 0 ) ; // Intensitaet blau
              osPaStream.write( (byte) 0 ) ; // Intensitaet gruen
              osPaStream.write( (byte) 0 ) ; // Intensitaet rot
              osPaStream.write( (byte) 0 ) ; // reserviert
        
              // 06 schwarz
              osPaStream.write( (byte) 0 ) ; // Intensitaet blau
              osPaStream.write( (byte) 0 ) ; // Intensitaet gruen
              osPaStream.write( (byte) 0 ) ; // Intensitaet rot
              osPaStream.write( (byte) 0 ) ; // reserviert
        
              // 07 schwarz
              osPaStream.write( (byte) 0 ) ; // Intensitaet blau
              osPaStream.write( (byte) 0 ) ; // Intensitaet gruen
              osPaStream.write( (byte) 0 ) ; // Intensitaet rot
              osPaStream.write( (byte) 0 ) ; // reserviert
        
              // 08 schwarz
              osPaStream.write( (byte) 0 ) ; // Intensitaet blau
              osPaStream.write( (byte) 0 ) ; // Intensitaet gruen
              osPaStream.write( (byte) 0 ) ; // Intensitaet rot
              osPaStream.write( (byte) 0 ) ; // reserviert
        
              // 09 schwarz
              osPaStream.write( (byte) 0 ) ; // Intensitaet blau
              osPaStream.write( (byte) 0 ) ; // Intensitaet gruen
              osPaStream.write( (byte) 0 ) ; // Intensitaet rot
              osPaStream.write( (byte) 0 ) ; // reserviert
        
              // 10 schwarz
              osPaStream.write( (byte) 0 ) ; // Intensitaet blau
              osPaStream.write( (byte) 0 ) ; // Intensitaet gruen
              osPaStream.write( (byte) 0 ) ; // Intensitaet rot
              osPaStream.write( (byte) 0 ) ; // reserviert
        
              // 11 schwarz
              osPaStream.write( (byte) 0 ) ; // Intensitaet blau
              osPaStream.write( (byte) 0 ) ; // Intensitaet gruen
              osPaStream.write( (byte) 0 ) ; // Intensitaet rot
              osPaStream.write( (byte) 0 ) ; // reserviert
        
              // 12 schwarz
              osPaStream.write( (byte) 0 ) ; // Intensitaet blau
              osPaStream.write( (byte) 0 ) ; // Intensitaet gruen
              osPaStream.write( (byte) 0 ) ; // Intensitaet rot
              osPaStream.write( (byte) 0 ) ; // reserviert
        
              // 13 schwarz
              osPaStream.write( (byte) 0 ) ; // Intensitaet blau
              osPaStream.write( (byte) 0 ) ; // Intensitaet gruen
              osPaStream.write( (byte) 0 ) ; // Intensitaet rot
              osPaStream.write( (byte) 0 ) ; // reserviert
        
              // 14 schwarz
              osPaStream.write( (byte) 0 ) ; // Intensitaet blau
              osPaStream.write( (byte) 0 ) ; // Intensitaet gruen
              osPaStream.write( (byte) 0 ) ; // Intensitaet rot
              osPaStream.write( (byte) 0 ) ; // reserviert
        
              // 15 schwarz
              osPaStream.write( (byte) 0 ) ; // Intensitaet blau
              osPaStream.write( (byte) 0 ) ; // Intensitaet gruen
              osPaStream.write( (byte) 0 ) ; // Intensitaet rot
              osPaStream.write( (byte) 0 ) ; // reserviert
        */

        // unterste Zeile wird zuerst gespeichert
        //for ( int iRow = 0 ; iRow < iClArr.length ; iRow++ )
        for (int iRow = (iClArr.length - 1); iRow >= 0; iRow--)
        {
            int iBytePerColCount = 0; // Zaehler Byte per Zeile
            for (int jCol = 0; jCol < iClArr[0].length; jCol++)
            {
                //osPaStream.write( iClArr[ iRow ][ jCol ] ) ;
                int iTmp = iClArr[iRow][jCol];
                BitmapUtil.writeInt24Intel(osPaStream, iTmp);
                iBytePerColCount += 3;
                //System.out.println( "" + ( ( iRow + 1 ) * ( jCol + 1 ) ) ) ;
            }
            // Auffuellen auf das naechste Vielfache von 4
            int iFuell = (iBytePerColCount % 4);
            if (iFuell > 0)
            {
                for (int i = iFuell; i < 4; i++)
                {
                    osPaStream.write((byte) 0);
                }
            }
        }
    }// end method writeBmpToStream

    /**
     *<BR>
     * Speichert das Sylk-Excel-Sheet als Datei<BR>
     * Verwenden Sie die Datei-Namens-Extension   .slk   oder  .xls
     *
     * @param strPaFileName Name und Pfad der Zieldatei als String
     *
     * @throws IOException
     */
    public void toBmpFile(String strPaFileName) throws IOException
    {
        File file = new File(strPaFileName);
        if (!file.exists())
        {
            file.createNewFile();
        }
        FileOutputStream fileOutputStream = new FileOutputStream(file);
        writeBmpToStream(fileOutputStream);
        fileOutputStream.close();
    } // end method toBmpFile

} // end class Bitmap24
