import java.io.*;

/**
 * Bitmap-Grafik (Microsoft-Windows)
 * mit 8 Bit per Pixel.
 * 
 * @author Heiner Kcker
 */
public class Bitmap8
{
    /*
       private final static String[][][] FONT =
          {
             {
                { " **** " ,
                  "*    *" ,
                  "******" ,
                  "*    *" ,
                  "*    *"
                }
                ,
                {
                   "*** " ,
                   "*  *" ,
                   "*  *" ,
                   "*** " ,
                   "*  *" ,
                   "*  *" ,
                   "*** "
                }
             }
          } ;
    */

    /**
     * Array mit den Pixeln jeweils als Byte-Elementartyp gespeichert
     */
    private byte[][] byClArr;

    /**
     * Anzahl Farben in der Palette
     */
    public final static int NUMBER_OF_COLORS = 16;

    /**
     * Farbcode weiss
     */
    public final static byte WHITE = (byte) 0;

    /**
     * Farbcode schwarz
     */
    public final static byte BLACK = (byte) 1;

    /**
     * Farbcode blau
     */
    public final static byte BLUE = (byte) 2;

    /**
     * Farbcode gruen
     */
    public final static byte GREEN = (byte) 3;

    /**
     * Farbcode rot
     */
    public final static byte RED = (byte) 4;

    /**
     * PALETTE als konstantes Array of Bytes
     */
    private final static byte[] PALETTE = {
        // Palette fuer 16 Farben
        // 00 weiss
         (byte) 0xFF, // Intensity blau
         (byte) 0xFF, // Intensity gruen
         (byte) 0xFF, // Intensity rot
         (byte) 0, // reserviert

        // 01 schwarz
         (byte) 0, // Intensity blau
         (byte) 0, // Intensity gruen
         (byte) 0, // Intensity rot
         (byte) 0, // reserviert

        // 02 blau
         (byte) 126, // Intensity blau
         (byte) 0, // Intensity gruen
         (byte) 0, // Intensity rot
         (byte) 0, // reserviert

        // 03 gruen
         (byte) 0, // Intensity blau
         (byte) 126, // Intensity gruen
         (byte) 0, // Intensity rot
         (byte) 0, // reserviert

        // 04 rot
         (byte) 0, // Intensity blau
         (byte) 0, // Intensity gruen
         (byte) 126, // Intensity rot
         (byte) 0, // reserviert

        // 05 schwarz
         (byte) 0, // Intensity blau
         (byte) 0, // Intensity gruen
         (byte) 0, // Intensity rot
         (byte) 0, // reserviert

        // 06 schwarz
         (byte) 0, // Intensity blau
         (byte) 0, // Intensity gruen
         (byte) 0, // Intensity rot
         (byte) 0, // reserviert

        // 07 schwarz
         (byte) 0, // Intensity blau
         (byte) 0, // Intensity gruen
         (byte) 0, // Intensity rot
         (byte) 0, // reserviert

        // 08 schwarz
         (byte) 0, // Intensity blau
         (byte) 0, // Intensity gruen
         (byte) 0, // Intensity rot
         (byte) 0, // reserviert

        // 09 schwarz
         (byte) 0, // Intensity blau
         (byte) 0, // Intensity gruen
         (byte) 0, // Intensity rot
         (byte) 0, // reserviert

        // 10 schwarz
         (byte) 0, // Intensity blau
         (byte) 0, // Intensity gruen
         (byte) 0, // Intensity rot
         (byte) 0, // reserviert

        // 11 schwarz
         (byte) 0, // Intensity blau
         (byte) 0, // Intensity gruen
         (byte) 0, // Intensity rot
         (byte) 0, // reserviert

        // 12 schwarz
         (byte) 0, // Intensity blau
         (byte) 0, // Intensity gruen
         (byte) 0, // Intensity rot
         (byte) 0, // reserviert

        // 13 schwarz
         (byte) 0, // Intensity blau
         (byte) 0, // Intensity gruen
         (byte) 0, // Intensity rot
         (byte) 0, // reserviert

        // 14 schwarz
         (byte) 0, // Intensity blau
         (byte) 0, // Intensity gruen
         (byte) 0, // Intensity rot
         (byte) 0, // reserviert

        // 15 schwarz
         (byte) 0, // Intensity blau
         (byte) 0, // Intensity gruen
         (byte) 0, // Intensity rot
         (byte) 0, // reserviert
    };

    /**
     * Konstruktor mit Grenangaben
     */
    public Bitmap8(int iPaRows, int iPaCols)
    {
        byClArr = new byte[iPaRows][iPaCols];
    } // end Konstruktor

    /**
     * Konstruktor mit Grenangaben und Hintergrundfarbe
     */
    public Bitmap8(int iPaRows, int iPaCols, byte pBackgroundColorByte)
    {
        byClArr = new byte[iPaRows][iPaCols];

        // Array mit Hintergrundfarbe setzen
        for (int iRow = 0; iRow < byClArr.length; iRow++)
        {
            for (int jCol = 0; jCol < byClArr[0].length; jCol++)
            {
                byClArr[iRow][jCol] = (byte) pBackgroundColorByte;
            }
        }
    } // end Konstruktor

    /**
     * Setzen eines Pixels mit einer bestimmten Farbe
     */
    public void setPixel(int iPaRow, int iPaCol, byte byPaColor)
    {
        if (iPaRow >= 0
            && iPaRow < byClArr.length
            && iPaCol >= 0
            && iPaCol < byClArr[iPaRow].length)
        {
            byClArr[iPaRow][iPaCol] = byPaColor;
        }
    } // end method setPixel

    /**
     * Linie zeichnen
     */
    public void drawLine(
        int iPaRow1,
        int iPaCol1,
        int iPaRow2,
        int iPaCol2,
        byte byPaColor)
    {
        //Prfung Parameter
        if (iPaRow1 >= 0
            && iPaRow1 < byClArr.length
            && // Zeile 1
        iPaRow2 >= 0
            && iPaRow2 < byClArr.length
            && // Zeile 2
        iPaCol1 >= 0
            && iPaCol1 < byClArr[0].length
            && // Spalte 1
        iPaCol2 >= 0
            && iPaCol2 < byClArr[0].length) // Spalte 2
        {
            if (iPaRow1 == iPaRow2)
            {
                // horizontale Linie
                //System.out.println( "horizontale Linie" );
                if (iPaCol1 > iPaCol2)
                    // Punkte mssen getauscht werden, sonst klappt es nicht
                {
                    int iTmp;
                    iTmp = iPaRow1;
                    iPaRow1 = iPaRow2;
                    iPaRow2 = iTmp;
                    iTmp = iPaCol1;
                    iPaCol1 = iPaCol2;
                    iPaCol2 = iTmp;
                }

                for (int i = iPaCol1; i <= iPaCol2; i++)
                {
                    //System.out.println( "Zeile " + iPaRow1 + " Spalte " + i );
                    // schneller als ber setPixel
                    byClArr[iPaRow1][i] = byPaColor;
                }
            }
            else if (iPaCol1 == iPaCol2)
            {
                // vertikale Linie
                //System.out.println( "vertikale Linie" );
                if (iPaRow1 > iPaRow2)
                    // Punkte mssen getauscht werden, sonst klappt es nicht
                {
                    int iTmp;
                    iTmp = iPaRow1;
                    iPaRow1 = iPaRow2;
                    iPaRow2 = iTmp;
                    iTmp = iPaCol1;
                    iPaCol1 = iPaCol2;
                    iPaCol2 = iTmp;
                }

                for (int i = iPaRow1; i <= iPaRow2; i++)
                {
                    // schneller als ber setPixel
                    byClArr[i][iPaCol1] = byPaColor;
                }
            }
            else if (Math.abs(iPaRow2 - iPaRow1) < Math.abs(iPaCol2 - iPaCol1))
            {
                //System.out.println( "flache Linie ( iPaRow2 - iPaRow1 ) " + ( iPaRow2 - iPaRow1 ) + " ( iPaCol2 - iPaCol1 ) " + ( iPaCol2 - iPaCol1 ) );
                // flache Linie
                if (iPaCol1 > iPaCol2)
                    // Punkte mssen getauscht werden, sonst klappt es nicht
                {
                    int iTmp;
                    iTmp = iPaRow1;
                    iPaRow1 = iPaRow2;
                    iPaRow2 = iTmp;
                    iTmp = iPaCol1;
                    iPaCol1 = iPaCol2;
                    iPaCol2 = iTmp;
                }

                // hier noch optmieren iStep gleich im if zuweisen
                float iStep =
                    ((float) iPaRow2 - (float) iPaRow1)
                        / ((float) iPaCol2 - (float) iPaCol1);

                float iRow = iPaRow1;
                for (int i = iPaCol1; i <= iPaCol2; i++)
                {
                    // schneller als ber setPixel
                    byClArr[Math.round(iRow)][i] = byPaColor;
                    iRow += iStep;
                }
            }
            else
            {
                // steile Linie
                //System.out.println( "steile Linie" );

                if (iPaRow1 > iPaRow2)
                    // Punkte mssen getauscht werden, sonst klappt es nicht
                {
                    int iTmp;
                    iTmp = iPaRow1;
                    iPaRow1 = iPaRow2;
                    iPaRow2 = iTmp;
                    iTmp = iPaCol1;
                    iPaCol1 = iPaCol2;
                    iPaCol2 = iTmp;
                }

                float iStep =
                    (((float) iPaCol2) - ((float) iPaCol1))
                        / (((float) iPaRow2) - ((float) iPaRow1));
                float iCol = iPaCol1;
                for (int i = iPaRow1; i <= iPaRow2; i++)
                {
                    //byClArr[ i ][ iCol ] = byPaColor ; // schneller als ber setPixel
                    setPixel(i, Math.round(iCol), byPaColor);
                    //System.out.println( "Zeile " + i + " Spalte " + Math.round( iCol ) );
                    iCol += iStep;
                }
            }
        }
    } // end method drawLine

    /**
     * Kreis Zeichnen
     *
     * @param iPaRow    Zeile Mittelpunkt
     * @param iPaCol    Spalte Mittelpunkt
     * @param iPaRadius Radius in Pixeln
     * @param byPaColor Farbe
     */
    public void drawCircle(
        int iPaRow,
        int iPaCol,
        int iPaRadius,
        byte byPaColor)
    {
        float fTmp;
        int iTmp;

        fTmp = (float) iPaRadius * (float) iPaRadius; // Quadrat

        for (int i = 0; i <= iPaRadius; i++)
        {
            iTmp =
                Math.round((float) Math.sqrt(fTmp - ((float) i * (float) i)));
            setPixel(iPaRow + iTmp, iPaCol + i, byPaColor);
            setPixel(iPaRow + iTmp, iPaCol - i, byPaColor);
            setPixel(iPaRow - iTmp, iPaCol + i, byPaColor);
            setPixel(iPaRow - iTmp, iPaCol - i, byPaColor);

            setPixel(iPaRow + i, iPaCol + iTmp, byPaColor);
            setPixel(iPaRow + i, iPaCol - iTmp, byPaColor);
            setPixel(iPaRow - i, iPaCol + iTmp, byPaColor);
            setPixel(iPaRow - i, iPaCol - iTmp, byPaColor);
        }
    } // end method drawCircle

    /**
     * Zeichnen eines Kreisabschnittes Pie. <BR>
     * gezeichnet wird im Kreis entgegen der Uhrzeigerrichtung
     *
     * @param iPaRow           Zeile Mittelpunkt
     * @param iPaCol           Spalte Mittelpunkt
     * @param iPaRadius        Radius in Pixeln
     * @param fPaAngleStart    Startwinkel in Bogenmass
     * @param fPaAngleEnd      Endwinkel in Bogenmass
     * @param byPaOutlineColor Farbe Ausssenlinie
     * @param bPaFill          Fllen ja/nein
     * @param byPaAreaColor    Farbe Flche
     */
    public void drawPie(
        int iPaRow,
        int iPaCol,
        int iPaRadius,
        float fPaAngleStart,
        float fPaAngleEnd,
        byte byPaOutlineColor,
        boolean bPaFill,
        byte byPaAreaColor)
    {
        // erst innen ausfllen und dann mit evtl. anderer Farbe den Umfang zeichnen
        if (bPaFill)
        {
            // wenn Fllen
            /*
                     // Punkt an Positition Radius-1 und Mittelwert aus Anfangs- und Endwinkel
                     int iTmp = iPaRadius - 1 ;
                     float fTmp = ( fPaAngleStart + fPaAngleEnd ) / 2 ; //
                     fillArea( iPaRow - Math.round( (float) Math.sin( fTmp ) * iTmp ) , iPaCol + Math.round( (float) Math.cos( fTmp ) * iTmp ) , byPaColor ) ;
                     //drawLine( iPaRow , iPaCol , iPaRow - Math.round( (float) Math.sin( fDrawAngle ) * iPaRadius ) , iPaCol + Math.round( (float) Math.cos( fDrawAngle ) * iPaRadius ) , byPaColor ) ;
            */
            // einen Kreis nach dem anderen Zeichnen
            for (float i = 1; i < iPaRadius; i += 0.2)
            {
                drawCircleSegment(
                    iPaRow,
                    iPaCol,
                    i,
                    fPaAngleStart,
                    fPaAngleEnd,
                    byPaAreaColor);
            }

        }
        // Zeichnen Anfangslinie
        drawLine(
            iPaRow,
            iPaCol,
            iPaRow - Math.round((float) Math.sin(fPaAngleStart) * iPaRadius),
            iPaCol + Math.round((float) Math.cos(fPaAngleStart) * iPaRadius),
            byPaOutlineColor);

        // Zeichnen Aussenbogen
        drawCircleSegment(
            iPaRow,
            iPaCol,
            (float) iPaRadius,
            fPaAngleStart,
            fPaAngleEnd,
            byPaOutlineColor);

        // Zeichnen Endlinie
        drawLine(
            iPaRow,
            iPaCol,
            iPaRow - Math.round((float) Math.sin(fPaAngleEnd) * iPaRadius),
            iPaCol + Math.round((float) Math.cos(fPaAngleEnd) * iPaRadius),
            byPaOutlineColor);
    } // end method drawPie

    /**
     * Zeichnen herausgehobenes Pie
     */
    public void drawExtendPie(
        int iPaRow,
        int iPaCol,
        int iPaRadius,
        float fPaAngleStart,
        float fPaAngleEnd,
        byte byPaOutlineColor,
        boolean bPaFill,
        byte byPaAreaColor,
        int iPaExtendRadius)
    {
        // Mittelpunkt um iPaExtendRadius in Richtung Mittelwert Anfangs- und Endwinkel verschieben
        float fAngelMedian = (fPaAngleStart + fPaAngleEnd) / 2;
        drawPie(
            iPaRow
                - Math.round((float) Math.sin(fAngelMedian) * iPaExtendRadius),
            iPaCol
                + Math.round((float) Math.cos(fAngelMedian) * iPaExtendRadius),
            iPaRadius,
            fPaAngleStart,
            fPaAngleEnd,
            byPaOutlineColor,
            bPaFill,
            byPaAreaColor);
        //System.out.println( "x " + iPaRow + " x " +  Math.round( ( (float) Math.sin( fAngelMedian ) / (float) Math.PI ) * iPaRadius ) + " y " + iPaCol + " y " + Math.round( ( (float) Math.cos( fAngelMedian ) / (float) Math.PI ) * iPaRadius ) ) ;
    } // end method drawExtendPie

    /**
     * Zeichnen Kreissegment
     */
    public void drawCircleSegment(
        int iPaRow,
        int iPaCol,
        float fPaRadius,
        float fPaAngleStart,
        float fPaAngleEnd,
        byte byPaColor)
    {
        // Iterieren ber Kreisbogen
        float fStepAngle = (float) Math.PI / fPaRadius / 4;
        // Winkelschrittweite ( soll kleiner sein als ein Pixel Ausssenlinie pro Schritt )
        float fDrawAngle;
        for (fDrawAngle = fPaAngleStart;
            fDrawAngle < fPaAngleEnd;
            fDrawAngle += fStepAngle)
        {
            setPixel(
                iPaRow - Math.round((float) Math.sin(fDrawAngle) * fPaRadius),
                iPaCol + Math.round((float) Math.cos(fDrawAngle) * fPaRadius),
                byPaColor);
        }
    } // end method drawCircleSegment

    /**
     * Zeichnen Pie-Chart in den Mittelpunkt der Bitmap
     * die Groesse wird automatisch ermittelt ( Radius = 1/4-Bildgroesse )
     * noch ohne Farbdefinitionen
     */
    public void pieChart(final double[] dArrPaChartValues)
    {
        //noch null-Prfung ChartArray

        int iRowCentre = byClArr.length / 2; // Hhe/2
        int iColCentre = byClArr[0].length / 2; // Breite/2
        int iRadius = Math.min(byClArr.length, byClArr[0].length) / 4;
        // Minimum Hhe/Breite / 4
        // Radius, um den die Kreisbgen nach aussen verschoben werden
        int iExtendRadius = 0; // iRadius / 10 ;

        // Summe aller Werte ermittteln
        double dChartSumme = 0;
        for (int i = 0; i < dArrPaChartValues.length; i++)
        {
            dChartSumme += dArrPaChartValues[i];
        }

        // Winkel auf Kreis aufteilen ( Gesamtkreis ist 2 * PI )
        double[] fArrSegmentWinkel = new double[dArrPaChartValues.length];
        for (int i = 0; i < dArrPaChartValues.length; i++)
        {
            // darf nicht kleiner Null sein
            fArrSegmentWinkel[i] =
                Math.max(0, dArrPaChartValues[i] / dChartSumme * 2 * Math.PI);
        }

        double dStartWinkel = 0;
        for (int i = 0; i < fArrSegmentWinkel.length; i++)
        {
            //drawPie( iRowCentre , iColCentre , iRadius , (float) dStartWinkel , (float) ( dStartWinkel + fArrSegmentWinkel[ i ] ) , BLACK , true , ( ( i % 2 ) == 0 ) ? RED : BLUE ) ;
            // Achtung Zuweisung dStartWinkel in Ausdruck
            drawExtendPie(
                iRowCentre,
                iColCentre,
                iRadius,
                (float) dStartWinkel,
                (float) (dStartWinkel += fArrSegmentWinkel[i]),
                BLACK,
                true,
                ((i % 2) == 0) ? RED : BLUE,
                iExtendRadius);
            //dStartWinkel += fArrSegmentWinkel[ i ] ;
        }
    } // end method pieChart

    /**
     * Ausgabe des Pixel-Arrays in einen OutputStream ( 8 Bit per Pixel )
     */
    public void writeBmpToStream(OutputStream osPaStream) throws IOException
    {
        // Komprimieren des gesamten Bildes im Speicher
        byte[] byArrCompress = compressRLE8();

        // File ID "BM" Dateikennung
        osPaStream.write('B');
        osPaStream.write('M');

        final int iBitmapWidth = byArrCompress.length;

        BitmapUtil.writeInt32Intel(osPaStream, 54 + iBitmapWidth); // Dateilaenge

        // vier reservierte Bytes
        osPaStream.write((byte) 0);
        osPaStream.write((byte) 0);
        osPaStream.write((byte) 0);
        osPaStream.write((byte) 0);

        // Startadresse Bitmap 54 entspricht 36H ( keine Palette )
        BitmapUtil.writeInt32Intel(osPaStream, 54 + PALETTE.length);

        // Groesse des Kopfes in Bytes, Position 0EH, Wert ist immer 40
        BitmapUtil.writeInt32Intel(osPaStream, 40);

        // Breite der Bitmap in Bildschirmpunkten
        BitmapUtil.writeInt32Intel(osPaStream, byClArr[0].length);

        // Hhe der Bitmap in Bildschirmpunkten
        BitmapUtil.writeInt32Intel(osPaStream, byClArr.length);

        // Anzahl der Farbebenen, muss 1 sein
        osPaStream.write((byte) 1);
        osPaStream.write((byte) 0);

        // Bits pro Pixel
        osPaStream.write((byte) 8);
        osPaStream.write((byte) 0);

        // Kennzeichnung fuer die Komprimierung RLE8 -> 1 ( Runlength-Encoding )
        osPaStream.write((byte) 1);
        osPaStream.write((byte) 0);
        osPaStream.write((byte) 0);
        osPaStream.write((byte) 0);

        // Groesse des Bitmaps in Bytes
        BitmapUtil.writeInt32Intel(osPaStream, iBitmapWidth);

        BitmapUtil.writeInt32Intel(osPaStream, 3780); // horizontale Aufloesung

        BitmapUtil.writeInt32Intel(osPaStream, 3780); // vertikale Aufloesung

        // Anzahl der Farben aus der Farbtabelle
        BitmapUtil.writeInt32Intel(osPaStream, NUMBER_OF_COLORS);

        // Anzahl wichtiger Farben  aus der Farbtabelle
        BitmapUtil.writeInt32Intel(osPaStream, NUMBER_OF_COLORS);

        // die gesamte Palette rausschreiben, hinterlegt als Konstante
        osPaStream.write(PALETTE);

        /*
        //Ausgeben ohne Kompression
              // unterste Zeile wird zuerst gespeichert
              for ( int iRow = ( byClArr.length - 1 ) ; iRow >= 0  ; iRow-- )
              {
                 int iBytePerColCount = 0 ; // Zaehler Byte per Zeile
                 for ( int jCol = 0 ; jCol < byClArr[ 0 ].length ; jCol++ )
                 {
                    osPaStream.write( byClArr[ iRow ][ jCol ] ) ;
                    iBytePerColCount++ ;
                 }
                 // Auffuellen auf das naechste Vielfache von 4
                 int iFuell = ( iBytePerColCount % 4 ) ;
                 if ( iFuell > 0 )
                 {
                    for ( int i = iFuell ; i < 4 ; i++ )
                    {
                       osPaStream.write( (byte) 0 ) ;
                    }
                 }
              }
        */
        /*
        funzt nicht wegen falscher Dateigroesse und Groesse Bilddatenbereich
              // Ausgeben mit Kompression RLE8
              // unterste Zeile wird zuerst gespeichert
              for ( int iRow = ( byClArr.length - 1 ) ; iRow >= 0  ; iRow-- )
              {
                 byte[] byArrTmp = compressRLE8( byClArr[ iRow ] ) ; // komprimiertes ByteArray liefern
        
                 int iBytePerColCount = 0 ; // Zaehler Byte per Zeile
                 for ( int jCol = 0 ; jCol < byArrTmp.length ; jCol++ )
                 {
                    osPaStream.write( byArrTmp[ jCol ] ) ;
                    iBytePerColCount++ ;
                 }
        
                 if ( iRow < ( byClArr.length - 1 ) )
                 {
                    // Delta-Record: Ende der Zeile
                    osPaStream.write( (byte) 0 ) ;
                    osPaStream.write( (byte) 0 ) ;
                    iBytePerColCount += 2 ;
        System.out.println( "Zeilenende" ) ;
                 }
                 else
                 {
                    // Delta-Record: Ende der Bitmap
                    osPaStream.write( (byte) 0 ) ;
                    osPaStream.write( (byte) 1 ) ;
        System.out.println( "Dateiende" ) ;
                 }
                 // Auffuellen auf das naechste Vielfache von 4
                 //int iFuell = ( iBytePerColCount % 4 ) ;
                 //if ( iFuell > 0 )
                 //{
                 //   for ( int i = iFuell ; i < 4 ; i++ )
                 //   {
                 //      osPaStream.write( (byte) 0 ) ;
                 //   }
                 //}
              }
        */
        osPaStream.write(byArrCompress);

    } // end method writeBmpToStream

    /**
     * Komprimieren der gesamten Bitmap 
     * in ein Byte-Array nach der RLE8-Methode
     * (Run-Length-Encoding).
     */
    private byte[] compressRLE8() throws IOException
    {
        ByteArrayOutputStream baosTmp = new ByteArrayOutputStream();

        // Komprimieren mit Kompression RLE8
        // unterste Zeile wird zuerst gespeichert
        for (int iRow = (byClArr.length - 1); iRow >= 0; iRow--)
        {
            // komprimiertes ByteArray liefern
            byte[] byArrTmp = BitmapUtil.compressLineRLE8(byClArr[iRow]);

//            for (int jCol = 0; jCol < byArrTmp.length; jCol++)
//            {
//                baosTmp.write(byArrTmp[jCol]);
//            }
            baosTmp.write( byArrTmp );

            if (iRow > 0)
                // Abfrage auf > 0 weil mit der letzten Zeile begonnen wird
            {
                // Delta-Record: Ende der Zeile
                baosTmp.write(0);
                baosTmp.write(0);
                //System.out.println( "Zeilenende " + iRow ) ;
            }
        }

        // Delta-Record: Ende der Bitmap
        baosTmp.write(0);
        baosTmp.write(1);
        //System.out.println( "Dateiende" ) ;

        return baosTmp.toByteArray();
    } // end method compressRLE8

    /**
     *<BR>
     * Speichert die Bitmap als Datei<BR>
     * Verwenden Sie die Datei-Namens-Extension   .bmp
     *
     * @param strPaFileName Name und Pfad der Zieldatei als String
     *
     * @throws IOException
     */
    public void toBmpFile(String strPaFileName) throws IOException
    {
        File file = new File(strPaFileName);
        if (!file.exists())
        {
            file.createNewFile();
        }
        FileOutputStream fileOutputStream = new FileOutputStream(file);
        writeBmpToStream(fileOutputStream);
        fileOutputStream.close();
    } // end method toBmpFile

    /**
     * nicht fertig.
     * 
     * Ausgabe des Pixel-Arrays in einen OutputStream als PNG-File
     *
     * hart codiert 24-Bit per Pixel plus Alphakanal
     */
    //   public void writePngToStream( OutputStream osPaStream )
    //      throws IOException
    //   {
    //      // Header
    //      //byte[] bArrHeaderPuff = new byte[ 13 ] ;
    //      ByteArrayOutputStream baosTmp = new ByteArrayOutputStream() ;
    //
    //      // PNG-Vorspann
    //      baosTmp.write( -119 ) ;
    //      baosTmp.write(   80 ) ;
    //      baosTmp.write(   78 ) ;
    //      baosTmp.write(   71 ) ;
    //      baosTmp.write(   13 ) ;
    //      baosTmp.write(   10 ) ;
    //      baosTmp.write(   26 ) ;
    //      baosTmp.write(   10 ) ;
    //
    //      // PNG-Header-Chunk
    //      writeInt32( baosTmp , 13 ) ;                  // Gre Header-Datenbereich immer 13
    //
    //      //baosTmp.write( "IHDR".getBytes() ) ;          // Kennzeichnung Header-Chunk
    //      baosTmp.write(   73 ) ;                       // I Kennzeichnung Header-Chunk "IHDR"
    //      baosTmp.write(   72 ) ;                       // H
    //      baosTmp.write(   68 ) ;                       // D
    //      baosTmp.write(   82 ) ;                       // R
    //
    //      writeInt32( baosTmp , byClArr[ 0 ].length ) ; // Breite der Bitmap in Bildschirmpunkten
    //      writeInt32( baosTmp , byClArr.length ) ;      // Hhe   der Bitmap in Bildschirmpunkten
    //      baosTmp.write( 8 ) ;                          // Bit pro Sample
    //      baosTmp.write( 6 ) ;                          // Colormodel 24 Bit + Alphakanal hart codiert
    //      baosTmp.write( 0 ) ;                          // Kompressionstyp
    //      baosTmp.write( 0 ) ;                          // Filter-Typ
    //      baosTmp.write( 0 ) ;                          // Interlace-Flag
    //
    //      CRC32 crc = new CRC32() ;
    //      crc.reset();
    //      byte[] bArrTmp = baosTmp.toByteArray() ;
    //      crc.update( bArrTmp , 8 , bArrTmp.length - 8 ) ; // Prfsumme ab 9. Byte ( Beginn Header )
    //
    //      long lCrcValue = crc.getValue();
    //      writeInt32( baosTmp , (int) lCrcValue ) ;
    //
    //      // rausschreiben
    //      osPaStream.write( baosTmp.toByteArray() ) ;
    //      // Ende PNG-Header-Chunk
    //
    //
    //
    //
    //      // Bilddatenbereich
    //
    //      int rowsLeft = byClArr.length ; // height;  // number of rows remaining to write
    //      int startRow = 0;       // starting row to process this time through
    //      int nRows;              // how many rows to grab at a time
    //
    //      byte[] scanLines;       // the scan lines to be compressed
    //      int scanPos;            // where we are in the scan lines
    //      //int startPos;           // where this line's actual pixels start (used for filtering)    momentan kein Filter
    //
    //      byte[] compressedLines; // the resultant compressed lines
    //      //int nCompressed;        // how big is the compressed area?
    //
    //      //int depth;              // color depth ( handle only 8 or 32 )
    //
    //      int bytesPerPixel = 4 ; // erst mal konstant 4 (encodeAlpha) ? 4 : 3;
    //
    //      Deflater scrunch = new Deflater( 9 ) ; //  9 fest verdrahtet    compressionLevel );
    //      ByteArrayOutputStream outBytes =
    //         new ByteArrayOutputStream(1024);
    //
    //      DeflaterOutputStream compBytes =
    //        new DeflaterOutputStream( outBytes, scrunch );
    //
    //      while (rowsLeft > 0)
    //      {
    //         //                          width
    //         nRows = Math.min( 32767 / ( byClArr[ 0 ].length * ( bytesPerPixel + 1 ) ) , rowsLeft ) ;
    //         // nRows = rowsLeft;
    //
    ///* nicht ntig
    //         int[] pixels = new int[width * nRows];
    //
    //          pg = new PixelGrabber(image, 0, startRow,
    //              width, nRows, pixels, 0, width);
    //          try {
    //              pg.grabPixels();
    //          }
    //          catch (Exception e) {
    //              System.err.println("interrupted waiting for pixels!");
    //              return false;
    //          }
    //          if ((pg.getStatus() & ImageObserver.ABORT) != 0) {
    //              System.err.println("image fetch aborted or errored");
    //              return false;
    //          }
    //*/
    //
    //
    //
    //          // Create a data chunk. scanLines adds "nRows" for the filter bytes.
    //          //                     width
    //          //scanLines = new byte[ byClArr[ 0 ].length * nRows * bytesPerPixel +  nRows ] ;
    //          scanLines = new byte[ byClArr[ 0 ].length * byClArr.length * bytesPerPixel ] ;
    //
    ///* erst mal ohne Filter
    //          if (filter == FILTER_SUB)
    //          {
    //              leftBytes = new byte[16];
    //          }
    //          if (filter == FILTER_UP)
    //          {
    //              priorRow = new byte[width*bytesPerPixel];
    //          }
    //*/
    //
    //          scanPos = 0;
    //          //startPos = 1; im Moment kein Filter
    //          for ( int i = 0 ; i < byClArr.length ; i++ )
    //          {
    //            for ( int j = 0 ; j  < byClArr[ 0 ].length ; j++ )
    //            {
    //
    ////System.out.println( "scanLines " + scanLines.length + " scanPos " + scanPos + " i " + i + " j " + j ) ;
    //
    //                 //       width
    //                 //if (i % byClArr[ 0 ].length == 0)  im Moment kein Filter
    //                 //{
    //                     // scanLines[scanPos++] = (byte) filter;
    //                     // startPos = scanPos;
    //                 //}
    //                 int iPixelOffset = byClArr[ i ][ j ] * 4 ; // Offset in PALETTE-Array
    //
    //if ( iPixelOffset + 2 > PALETTE.length )
    //{
    //   iPixelOffset = 0 ; // abfangen Array-berschreitung
    //}
    //
    //                 scanLines[ scanPos++ ] = (byte) ( PALETTE[ iPixelOffset + 2 ] ) ; // red
    //                 scanLines[ scanPos++ ] = (byte) ( PALETTE[ iPixelOffset + 1 ] ) ; // green
    //                 scanLines[ scanPos++ ] = (byte) ( PALETTE[ iPixelOffset     ] ) ; // blue
    //
    //                 //if (encodeAlpha) fest eingeschalten
    //                 //{
    //                 scanLines[scanPos++] = (byte) ( iPixelOffset > 0 ? 0xff : 0 ) ; // bei Farbe null durchsichtig, sonst undurchsichtig
    //                 //}
    //
    ///* erst mal ohne Filter
    //              if ((i % width == width-1) && (filter != FILTER_NONE))
    //              {
    //                  if (filter == FILTER_SUB)
    //                  {
    //                      filterSub( scanLines, startPos, width );
    //                  }
    //                  if (filter == FILTER_UP)
    //                  {
    //                      filterUp( scanLines, startPos, width );
    //                  }
    //              }
    //*/
    //            }
    //         }
    //
    //
    //      // Write these lines to the output area
    //          compBytes.write( scanLines, 0, scanPos );
    //
    //
    //          startRow += nRows;
    //          rowsLeft -= nRows;
    //      }
    //
    ////System.out.println( "Schleifenende" ) ;
    //
    //      compBytes.close();
    //
    //      // Write the compressed bytes
    //      compressedLines = outBytes.toByteArray();
    //      //nCompressed = compressedLines.length;
    //
    //      crc.reset();
    //
    //      //osPaStream.write( "IDAT".getBytes() ) ;          // Kennzeichnung Daten-Chunk
    //      osPaStream.write(   73 ) ;                       // I Kennzeichnung Daten-Chunk "IDAT"
    //      osPaStream.write(   68 ) ;                       // D
    //      osPaStream.write(   65 ) ;                       // A
    //      osPaStream.write(   84 ) ;                       // T
    //
    //      osPaStream.write( compressedLines ) ;
    //
    //      crc.update("IDAT".getBytes());
    //
    //      osPaStream.write( compressedLines ) ;
    //      crc.update( compressedLines, 0,  compressedLines.length );
    //
    //      lCrcValue = crc.getValue();
    //      writeInt32( osPaStream , (int) lCrcValue );
    //      scrunch.finish();
    //
    //
    //
    //
    //      // Ende Bilddatenbereich
    //
    //
    //
    //
    //
    //
    //
    //      // PNG-"IEND"-Chunk  ( wird direkt in den als Parameter bergebenen OutputStram gesendet )
    //      // Lnge Datenbereich des Chunk
    //      osPaStream.write( 0 ) ;
    //      osPaStream.write( 0 ) ;
    //      osPaStream.write( 0 ) ;
    //      osPaStream.write( 0 ) ;
    //      osPaStream.write(   73 ) ;                       // I Kennzeichnung End-Chunk "IEND"
    //      osPaStream.write(   69 ) ;                       // E
    //      osPaStream.write(   78 ) ;                       // N
    //      osPaStream.write(   68 ) ;                       // D
    //
    //      crc.reset();
    //      crc.update("IEND".getBytes());
    //      lCrcValue = crc.getValue();
    //      writeInt32( osPaStream , (int) lCrcValue ) ; // hier mal eine Konstante anlegen
    //      // Ende PNG-"IEND"-Chunk
    //   } // end method writePngToStream

    /**
     * nicht fertig.
     * 
     *<BR>
     * Speichert die Bitmap als Datei . <BR>
     * Verwenden Sie die Datei-Namens-Extension '*.png' 
     *
     * @param strPaFileName Name und Pfad der Zieldatei als String
     *
     * @throws IOException
     */
    //  public void toPngFile( String strPaFileName )
    //      throws IOException
    //   {
    //      File file = new File( strPaFileName ) ;
    //      if ( ! file.exists() )
    //      {
    //         file.createNewFile() ;
    //      }
    //      FileOutputStream fileOutputStream = new FileOutputStream( file ) ;
    //      writePngToStream( fileOutputStream ) ;
    //      fileOutputStream.close() ;
    //   } // end method toPngFile

} // end class Bitmap8