import java.io.*;

/**
 * @author Heiner kcker
 */
public class BitmapUtil
{

    /**
     * Konstructor private, instanziieren verboten.
     * 
     */
    private BitmapUtil()
    {
        super();
    }

    /**
     * Schreiben eines 16-bit-Integer in Intel-Notation auf ein OutputStream
     */
    public static final void writeInt16Intel(
        final OutputStream osPaStream,
        int iPaValue)
        throws IOException
    {
        osPaStream.write((byte) (iPaValue % 256));
        iPaValue = iPaValue >>> 8;
        osPaStream.write((byte) (iPaValue % 256));
        iPaValue = iPaValue >>> 8;
    } // end method writeInt16Intel

    /**
     * Schreiben eines int-Wertes als 16-Bit-Wert auf einen OutputStream 
     * ( nicht im Intel-Format )
     */
    public static final void writeInt16(
        final OutputStream osPaStream,
        final int iPaValue)
        throws IOException
    {
        osPaStream.write((byte) (iPaValue >> 8 & 0xff));
        osPaStream.write((byte) (iPaValue & 0xff));
    }

    /**
     * Schreiben eines 24-bit-Integer in Intel-Notation auf ein OutputStream
     */
    public static final void writeInt24Intel(
        final OutputStream pOutputStream,
        int pIntValue)
        throws IOException
    {
        pOutputStream.write((byte) (pIntValue % 256));
        pIntValue = pIntValue >>> 8;
        pOutputStream.write((byte) (pIntValue % 256));
        pIntValue = pIntValue >>> 8;
        pOutputStream.write((byte) (pIntValue % 256));
    }

    /**
     * Schreiben eines 32-bit-Integer in Intel-Notation auf ein OutputStream
     */
    public static final void writeInt32Intel(
        final OutputStream osPaStream,
        int iPaValue)
        throws IOException
    {
        osPaStream.write((byte) (iPaValue % 256));
        iPaValue = iPaValue >>> 8;
        osPaStream.write((byte) (iPaValue % 256));
        iPaValue = iPaValue >>> 8;
        osPaStream.write((byte) (iPaValue % 256));
        iPaValue = iPaValue >>> 8;
        osPaStream.write((byte) (iPaValue % 256));
    } // end method writeInt32Intel

    /**
     * Schreiben eines int-Wertes als 32-Bit-Wert auf einen OutputStream 
     * ( nicht im Intel-Format )
     */
    public static final void writeInt32(
        final OutputStream osPaStream,
        final int iPaValue)
        throws IOException
    {
        osPaStream.write((byte) (iPaValue >> 24 & 0xff));
        osPaStream.write((byte) (iPaValue >> 16 & 0xff));
        osPaStream.write((byte) (iPaValue >> 8 & 0xff));
        osPaStream.write((byte) (iPaValue & 0xff));
    }

    /**
     * Komprimiert eine Zeile in RLE8-Methode ( Run-length-Encoding )
     * ganz einfache Version
     */
    public static byte[] compressLineRLE8(final byte[] byPaArr)
    {
        int iSrcPtr = 0; // Zeiger auf die aktuelle Position im Quell-Array
        int iDstPtr = 0; // Zeiger auf die aktuelle Position im Ziel-Array
        byte[] byArrTmp = new byte[byPaArr.length * 2]; // Ziel-Array

        // Durchlauf ber alles
        for (iSrcPtr = 0; iSrcPtr < byPaArr.length;)
        {
            // Durchlauf je gleichen Byte, aber maximal 255
            byte bTmp = byPaArr[iSrcPtr];
            int iTmpScrPtr = iSrcPtr + 1; // Zeiger immer auf das nchste Byte
            int iAnzahl = 0;
            for (iAnzahl = 1;
                iTmpScrPtr < byPaArr.length
                    && iAnzahl < 255
                    && bTmp == byPaArr[iTmpScrPtr];
                iAnzahl++)
            {
                iTmpScrPtr++; // inkrementieren Zeiger auf nchste Position
            }

            // Anzahl ( konvertieren in vorzeichenloses Byte )
            byArrTmp[iDstPtr] = (byte) (iAnzahl & 0xFF);
            iDstPtr++;
            byArrTmp[iDstPtr] = bTmp; // Farbwert ( eigentliches Datenbyte )
            iDstPtr++;

            iSrcPtr += iAnzahl; // um Anzahl weiterschalten
            //System.out.println( "iDstPtr " + iDstPtr + " iSrcPtr " + iSrcPtr ) ;
        }

        //System.out.println( "Umladen von byArrTmp nach byArrReturn mit Setzen Array-Groesse" ) ;

        // Umladen von byArrTmp nach byArrReturn mit Setzen Array-Groesse
        byte[] byArrReturn = new byte[iDstPtr];
        for (int i = 0; i < byArrReturn.length; i++)
        {
            byArrReturn[i] = byArrTmp[i];
        }

        return byArrReturn;
    } // end method compressLineRLE8

}
